@@ -22,7 +22,7 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with ldap3 in the COPYING and COPYING.LESSER files.
 # If not, see <http://www.gnu.org/licenses/>.
-from copy import deepcopy
+from copy import deepcopy, copy
 from os import linesep
 from threading import RLock, Lock
 from functools import reduce
@@ -30,9 +30,9 @@ import json
 
 from .. import ANONYMOUS, SIMPLE, SASL, MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE, get_config_parameter, DEREF_ALWAYS, \
     SUBTREE, ASYNC, SYNC, NO_ATTRIBUTES, ALL_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES, MODIFY_INCREMENT, LDIF, ASYNC_STREAM, \
-    RESTARTABLE, ROUND_ROBIN, REUSABLE, AUTO_BIND_DEFAULT, AUTO_BIND_NONE, AUTO_BIND_TLS_BEFORE_BIND,\
+    RESTARTABLE, ROUND_ROBIN, REUSABLE, AUTO_BIND_DEFAULT, AUTO_BIND_NONE, AUTO_BIND_TLS_BEFORE_BIND, SAFE_SYNC, \
     AUTO_BIND_TLS_AFTER_BIND, AUTO_BIND_NO_TLS, STRING_TYPES, SEQUENCE_TYPES, MOCK_SYNC, MOCK_ASYNC, NTLM, EXTERNAL,\
-    DIGEST_MD5, GSSAPI, PLAIN
+    DIGEST_MD5, GSSAPI, PLAIN, DSA, SCHEMA, ALL
 
 from .results import RESULT_SUCCESS, RESULT_COMPARE_TRUE, RESULT_COMPARE_FALSE
 from ..extend import ExtendedOperationsRoot
@@ -52,6 +52,7 @@ from ..protocol.sasl.digestMd5 import sasl_digest_md5
 from ..protocol.sasl.external import sasl_external
 from ..protocol.sasl.plain import sasl_plain
 from ..strategy.sync import SyncStrategy
+from ..strategy.safeSync import SafeSyncStrategy
 from ..strategy.mockAsync import MockAsyncStrategy
 from ..strategy.asynchronous import AsyncStrategy
 from ..strategy.reusable import ReusableStrategy
@@ -65,8 +66,7 @@ from .usage import ConnectionUsage
 from .tls import Tls
 from .exceptions import LDAPUnknownStrategyError, LDAPBindError, LDAPUnknownAuthenticationMethodError, \
     LDAPSASLMechanismNotSupportedError, LDAPObjectClassError, LDAPConnectionIsReadOnlyError, LDAPChangeError, LDAPExceptionError, \
-    LDAPObjectError, LDAPSocketReceiveError, LDAPAttributeError, LDAPInvalidValueError, LDAPConfigurationError, \
-    LDAPInvalidPortError
+    LDAPObjectError, LDAPSocketReceiveError, LDAPAttributeError, LDAPInvalidValueError, LDAPInvalidPortError, LDAPStartTLSError
 
 from ..utils.conv import escape_bytes, prepare_for_stream, check_json_dict, format_json, to_unicode
 from ..utils.log import log, log_enabled, ERROR, BASIC, PROTOCOL, EXTENDED, get_library_log_hide_sensitive_data
@@ -80,6 +80,7 @@ SASL_AVAILABLE_MECHANISMS = [EXTERNAL,
                              PLAIN]
 
 CLIENT_STRATEGIES = [SYNC,
+                     SAFE_SYNC,
                      ASYNC,
                      LDIF,
                      RESTARTABLE,
@@ -116,7 +117,6 @@ def _format_socket_endpoints(sock):
     return '<no socket>'
 
 
-# noinspection PyProtectedMember
 class Connection(object):
     """Main ldap connection class.
 
@@ -181,7 +181,6 @@ class Connection(object):
     :param source_port_list: a list of source ports to choose from when opening the connection to the server. Cannot be specified with source_port
     :type source_port_list: list
     """
-
     def __init__(self,
                  server,
                  user=None,
@@ -323,6 +322,8 @@ class Connection(object):
 
             if self.strategy_type == SYNC:
                 self.strategy = SyncStrategy(self)
+            elif self.strategy_type == SAFE_SYNC:
+                self.strategy = SafeSyncStrategy(self)
             elif self.strategy_type == ASYNC:
                 self.strategy = AsyncStrategy(self)
             elif self.strategy_type == LDIF:
@@ -352,7 +353,7 @@ class Connection(object):
             self.post_send_search = self.strategy.post_send_search
 
             if not self.strategy.no_real_dsa:
-                self.do_auto_bind()
+                self._do_auto_bind()
             # else:  # for strategies with a fake server set get_info to NONE if server hasn't a schema
             #     if self.server and not self.server.schema:
             #         self.server.get_info = NONE
@@ -362,7 +363,16 @@ class Connection(object):
                 else:
                     log(BASIC, 'instantiated Connection: <%r>', self)
 
-    def do_auto_bind(self):
+    def _prepare_return_value(self, status, response=False):
+        if self.strategy.thread_safe:
+            temp_response = self.response
+            self.response = None
+            temp_request = self.request
+            self.request = None
+            return status, deepcopy(self.result), deepcopy(temp_response) if response else None, copy(temp_request)
+        return status
+
+    def _do_auto_bind(self):
         if self.auto_bind and self.auto_bind not in [AUTO_BIND_NONE, AUTO_BIND_DEFAULT]:
             if log_enabled(BASIC):
                 log(BASIC, 'performing automatic bind for <%s>', self)
@@ -371,17 +381,28 @@ class Connection(object):
             if self.auto_bind == AUTO_BIND_NO_TLS:
                 self.bind(read_server_info=True)
             elif self.auto_bind == AUTO_BIND_TLS_BEFORE_BIND:
-                self.start_tls(read_server_info=False)
-                self.bind(read_server_info=True)
+                if self.start_tls(read_server_info=False):
+                    self.bind(read_server_info=True)
+                else:
+                    error = 'automatic start_tls befored bind not successful' + (' - ' + self.last_error if self.last_error else '')
+                    if log_enabled(ERROR):
+                        log(ERROR, '%s for <%s>', error, self)
+                    self.unbind()  # unbind anyway to close connection
+                    raise LDAPStartTLSError(error)
             elif self.auto_bind == AUTO_BIND_TLS_AFTER_BIND:
                 self.bind(read_server_info=False)
-                self.start_tls(read_server_info=True)
+                if not self.start_tls(read_server_info=True):
+                    error = 'automatic start_tls after bind not successful' + (' - ' + self.last_error if self.last_error else '')
+                    if log_enabled(ERROR):
+                        log(ERROR, '%s for <%s>', error, self)
+                    self.unbind()
+                    raise LDAPStartTLSError(error)
             if not self.bound:
-                self.last_error = 'automatic bind not successful' + (' - ' + self.last_error if self.last_error else '')
+                error = 'automatic bind not successful' + (' - ' + self.last_error if self.last_error else '')
                 if log_enabled(ERROR):
-                    log(ERROR, '%s for <%s>', self.last_error, self)
+                    log(ERROR, '%s for <%s>', error, self)
                 self.unbind()
-                raise LDAPBindError(self.last_error)
+                raise LDAPBindError(error)
 
     def __str__(self):
         s = [
@@ -513,7 +534,6 @@ class Connection(object):
 
             return self
 
-    # noinspection PyUnusedLocal
     def __exit__(self, exc_type, exc_val, exc_tb):
         with self.connection_lock:
             context_bound, context_closed = self._context_state.pop()
@@ -646,7 +666,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done BIND operation, result <%s>', self.bound)
 
-            return self.bound
+            return self._prepare_return_value(self.bound, self.result)
 
     def rebind(self,
                user=None,
@@ -692,7 +712,7 @@ class Connection(object):
                     raise LDAPBindError('Unable to rebind as a different user, furthermore the server abruptly closed the connection')
             else:
                 self.strategy.pool.rebind_pool()
-                return True
+                return self._prepare_return_value(True, self.result)
 
     def unbind(self,
                controls=None):
@@ -724,7 +744,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done UNBIND operation, result <%s>', True)
 
-            return True
+            return self._prepare_return_value(True)
 
     def search(self,
                search_base,
@@ -838,7 +858,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done SEARCH operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value, response=True)
 
     def compare(self,
                 dn,
@@ -892,7 +912,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done COMPARE operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def add(self,
             dn,
@@ -981,7 +1001,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done ADD operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def delete(self,
                dn,
@@ -1025,7 +1045,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done DELETE operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def modify(self,
                dn,
@@ -1115,7 +1135,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done MODIFY operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def modify_dn(self,
                   dn,
@@ -1172,7 +1192,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done MODIFY DN operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def abandon(self,
                 message_id,
@@ -1205,7 +1225,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done ABANDON operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def extended(self,
                  request_name,
@@ -1239,15 +1259,16 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done EXTENDED operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value, response=True)
 
     def start_tls(self, read_server_info=True):  # as per RFC4511. Removal of TLS is defined as MAY in RFC4511 so the client can't implement a generic stop_tls method0
-
         if log_enabled(BASIC):
             log(BASIC, 'start START TLS operation via <%s>', self)
 
         with self.connection_lock:
             return_value = False
+            self.result = None
+
             if not self.server.tls:
                 self.server.tls = Tls()
 
@@ -1271,7 +1292,7 @@ class Connection(object):
             if log_enabled(BASIC):
                 log(BASIC, 'done START TLS operation, result <%s>', return_value)
 
-            return return_value
+            return self._prepare_return_value(return_value)
 
     def do_sasl_bind(self,
                      controls):
@@ -1474,7 +1495,8 @@ class Connection(object):
                 target.writelines(self.response_to_json(raw=raw, indent=indent, sort=sort))
                 target.close()
 
-    def _fire_deferred(self, read_info=True):
+    def _fire_deferred(self, read_info=None):
+        # if read_info is None reads the schema and server info if not present, if False doesn't read server info, if True reads always server info
         with self.connection_lock:
             if self.lazy and not self._executing_deferred:
                 self._executing_deferred = True
@@ -1485,10 +1507,15 @@ class Connection(object):
                     if self._deferred_open:
                         self.open(read_server_info=False)
                     if self._deferred_start_tls:
-                        self.start_tls(read_server_info=False)
+                        if not self.start_tls(read_server_info=False):
+                            error = 'deferred start_tls not successful' + (' - ' + self.last_error if self.last_error else '')
+                            if log_enabled(ERROR):
+                                log(ERROR, '%s for <%s>', error, self)
+                            self.unbind()
+                            raise LDAPStartTLSError(error)
                     if self._deferred_bind:
                         self.bind(read_server_info=False, controls=self._bind_controls)
-                    if read_info:
+                    if (read_info is None and (not self.server.info and self.server.get_info in [DSA, ALL]) or (not self.server.schema and self.server.get_info in [SCHEMA, ALL])) or read_info:
                         self.refresh_server_info()
                 except LDAPExceptionError as e:
                     if log_enabled(ERROR):
@@ -1501,10 +1528,10 @@ class Connection(object):
     def entries(self):
         if self.response:
             if not self._entries:
-                self._entries = self._get_entries(self.response)
+                self._entries = self._get_entries(self.response, self.request)
         return self._entries
 
-    def _get_entries(self, search_response):
+    def _get_entries(self, search_response, search_request):
         with self.connection_lock:
             from .. import ObjectDef, Reader
 
@@ -1529,7 +1556,7 @@ class Connection(object):
                 object_def += list(attr_set)  # converts the set in a list to be added to the object definition
                 object_defs.append((attr_set,
                                     object_def,
-                                    Reader(self, object_def, self.request['base'], self.request['filter'], attributes=attr_set) if self.strategy.sync else Reader(self, object_def, '', '', attributes=attr_set))
+                                    Reader(self, object_def, search_request['base'], search_request['filter'], attributes=attr_set) if self.strategy.sync else Reader(self, object_def, '', '', attributes=attr_set))
                                    )  # objects_defs contains a tuple with the set, the ObjectDef and a cursor
 
             entries = []
